@@ -2,7 +2,7 @@ package App::cpanel;
 
 use Exporter 'import';
 
-our $VERSION = '0.003';
+our $VERSION = '0.004';
 our @EXPORT_OK = qw(dispatch_cmd_print dispatch_cmd_raw_p dir_walk_p);
 
 =head1 NAME
@@ -35,6 +35,8 @@ App::cpanel - CLI for cPanel UAPI and API 2
 
   # download
   $ cpanel mirror public_html public_html cpanel localfs
+  # upload
+  $ cpanel mirror public_html public_html localfs cpanel
 
 =head1 DESCRIPTION
 
@@ -67,14 +69,14 @@ maps. Assumes UNIX-like semantics in filenames, i.e. C<$dir/$file>.
 
 Returns a promise of completion.
 
-The maps are hash-refs whose values are functions.
+The maps are hash-refs whose values are functions, and the keys are:
 
 =head3 ls
 
 Takes C<$dir>. Returns a promise of two hash-refs, of directories and of
 files. Each has keys of relative filename, values are an array-ref
 containing a string octal number representing UNIX permissions, and a
-number giving the C<mtime>.
+number giving the C<mtime>. Must reject if does not exist.
 
 =head3 mkdir
 
@@ -116,7 +118,7 @@ use warnings;
 use Mojo::URL;
 use Mojo::UserAgent;
 use Mojo::File qw(path);
-use Mojo::Util qw(dumper);
+use Mojo::Util qw(dumper encode);
 
 my %cmd2func = (
   uapi => [ \&uapi_p, 1 ],
@@ -152,7 +154,7 @@ sub dispatch_cmd_print {
   die "Unknown command '$cmd'\n" unless my $info = $cmd2func{$cmd};
   my $p = dispatch_cmd_raw_p($cmd, @_);
   $p = $p->then(\&dumper) if $info->[1];
-  $p->then(sub { print @_ }, sub { warn @_ })->wait;
+  $p->then(sub { print @_ }, sub { warn encode 'UTF-8', join '', @_ })->wait;
 }
 
 sub dispatch_cmd_raw_p {
@@ -191,6 +193,21 @@ sub read_file {
 sub read_token { read_file($token_file) }
 sub read_domain { read_file($domain_file) }
 
+sub _error_or_json {
+  my $res = $_[0]->res;
+  die $res->code . ": " . $res->message . "\n" if $res->code != 200;
+  $res->json;
+}
+
+sub _uapi_error_or_json {
+  my $json = $_[0];
+  if (!$json->{status}) {
+    die join '', "Failed:\n",
+      map "$_\n", map @{ $json->{$_} || [] }, qw(errors warnings);
+  }
+  $json;
+}
+
 sub uapi_p {
   my ($module, $function, @args) = @_;
   die "No module\n" unless $module;
@@ -203,7 +220,7 @@ sub uapi_p {
   my $tx_p = api_request 'get_p', $domain, $token,
     [ 'execute', $module, $function ],
     $args_hash;
-  $tx_p->then(sub { $_[0]->res->json });
+  $tx_p->then(\&_error_or_json)->then(\&_uapi_error_or_json);
 }
 
 sub download_p {
@@ -213,7 +230,11 @@ sub download_p {
   my $tx_p = api_request 'get_p', $domain, $token,
     [ 'download' ],
     { skipencode => 1, file => $file };
-  $tx_p->then(sub { $_[0]->res->body });
+  $tx_p->then(sub {
+    my $res = $_[0]->res;
+    die $res->code . ": " . $res->message . "\n" if $res->code != 200;
+    $res->body;
+  });
 }
 
 sub make_upload_form {
@@ -241,7 +262,20 @@ sub upload_p {
     undef,
     form => make_upload_form($dir, @files),
     ;
-  $tx_p->then(sub { $_[0]->res->json });
+  $tx_p->then(\&_error_or_json)->then(\&_uapi_error_or_json);
+}
+
+sub _api2_error_or_json {
+  my $json = $_[0];
+  my $result = $json->{cpanelresult};
+  if (!$result or !$result->{event}{result} or $result->{error}) {
+    die join '', "Failed:\n",
+      map "$_\n",
+      ($result->{error} ? $result->{error} : ()),
+      (map "$_->{src}: $_->{err}", grep !$_->{result}, @{$result->{data} || []}),
+      ;
+  }
+  $json;
 }
 
 sub api2_p {
@@ -261,7 +295,7 @@ sub api2_p {
       cpanel_jsonapi_apiversion => 2,
       %{ $args_hash || {} },
     };
-  $tx_p->then(sub { $_[0]->res->json });
+  $tx_p->then(\&_error_or_json)->then(\&_api2_error_or_json);
 }
 
 sub dir_walk_p {
@@ -271,11 +305,14 @@ sub dir_walk_p {
     $to_dir_create_p = Mojo::Promise->resolve(1);
   } else {
     my $from_dir_perms;
-    $to_dir_create_p = $from_map->{ls}->(path($from_dir)->dirname)->then(sub {
+    $to_dir_create_p = $to_map->{ls}->($to_dir)->catch(sub {
+      # only create if ls fails
+      $to_map->{mkdir}->($to_dir)
+    })->then(sub {
+      $from_map->{ls}->(path($from_dir)->dirname)
+    })->then(sub {
       my ($dirs, $files) = @_;
       $from_dir_perms = $dirs->{path($from_dir)->basename}[0] || '0755';
-    })->then(sub {
-      $to_map->{mkdir}->($to_dir)
     })->then(sub {
       $to_map->{chmod}->($to_dir, $from_dir_perms)
     });
@@ -315,10 +352,10 @@ sub localfs_ls {
   my ($dir) = @_;
   my $dir_path = path($dir);
   my %files = map {
-    ($_->basename => [ sprintf("%o", $_->lstat->mode), $_->lstat->mtime ])
+    ($_->basename => [ sprintf("%04o", $_->lstat->mode & 07777), $_->lstat->mtime ])
   } $dir_path->list->each;
   my %dirs = map {
-    ($_->basename => [ sprintf("%o", $_->lstat->mode), $_->lstat->mtime ])
+    ($_->basename => [ sprintf("%04o", $_->lstat->mode & 07777), $_->lstat->mtime ])
   } $dir_path->list({dir => 1})->grep(sub { !$files{$_->basename} })->each;
   Mojo::Promise->resolve(\%dirs, \%files);
 }
@@ -369,19 +406,19 @@ sub cpanel_read {
 sub cpanel_mkdir {
   my ($dir) = @_;
   $dir = path $dir;
-  uapi_p qw(Fileman mkdir), { path => $dir->dirname, name => $dir->basename };
+  api2_p qw(Fileman mkdir), { path => $dir->dirname, name => $dir->basename };
 }
 
 sub cpanel_write {
   my ($dir, $file, $content) = @_;
-  uapi_p qw(Fileman savefile), {
+  api2_p qw(Fileman savefile), {
     dir => $dir, filename => $file, content => $content,
   };
 }
 
 sub cpanel_chmod {
   my ($path, $perms) = @_;
-  uapi_p qw(Fileman fileop), {
+  api2_p qw(Fileman fileop), {
     op => 'chmod', metadata => $perms, sourcefiles => $path,
   };
 }
